/************************************************************************
* Copyright (c) 1995 Valentin Shopov, Blagoevgrad, Bulgaria		*
*									*
* Permission to use, copy, modify, and distribute this material for any	*
* purpose and without fee is hereby granted, provided that the above	*
* copyright notice and this permission notice appear in all copies, and	*
* that the name of the author not be used in advertising or publicity 	*
* pertaining to this material without the specific, prior written 	*
* permission of the author.						*
*									*
* THE AUTHOR MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR SUITABILITY	*  
* OF THIS MATERIAL FOR ANY PURPOSE. IT IS PROVIDED "AS IS", WITHOUT ANY	* 
* EXPRESS OR IMPLIED WARRANTIES.					*
*************************************************************************/

/**************************ciilib2.0.2***********************************
 *                                                                      *
 *        Character to Inline Image Conversion Library (CIILIB)         *
 *                                                                      *
 *         x16 and x14 BDF Font to CIILIB XBM source convertor.		*
 *              	by Valentin Shopov		                *
 *			valentin@rs.aubg.bg				*
 ************************************************************************/





#include <stdio.h>
#include <string.h>
#include <memory.h>
#include "cii.h"

char curcol[2048], currow[2048];
int KANIND[0xffff];
static char *info = "bdf2xbm.inf";

int ChCMax = -1;
struct CharCodesXBM ChC[256];
struct CharCodesXBM SupChC[] =
{
  {
    "jp",
    "JISX0208",
    NULL, NULL,
    0, 0, 0, 0, 0, 0
  }
  ,
  {
    "ksp",
    "JISX0212",
    NULL, NULL,
    0, 0, 0, 0, 0, 0
  }
  ,
  {
    "rk",
    "JISX0201",
    NULL, NULL,
    0, 0, 0, 0, 0, 0
  }
  ,
  {
    "gb",
    "GB2312",
    NULL, NULL,
    0, 0, 0, 0, 0, 0
  }
  ,
  {
    "big5",
    "BIG5",
    NULL, NULL,
    0, 0, 0, 0, 0, 0
  }
  ,
  {
    "kr",
    "KSC5601",
    NULL, NULL,
    0, 0, 0, 0, 0, 0
  }
  ,
  {
    "ltn1",
    "ISO8859-1",
    NULL, NULL,
    0, 0, 0, 0, 0, 0
  }
  ,
  {
    "ltn2",
    "ISO8859-2",
    NULL, NULL,
    0, 0, 0, 0, 0, 0
  }
  ,
  {
    "ltn3",
    "ISO8859-3",
    NULL, NULL,
    0, 0, 0, 0, 0, 0
  }
  ,
  {
    "ltn4",
    "ISO8859-4",
    NULL, NULL,
    0, 0, 0, 0, 0, 0
  }
  ,
  {
    "cyrillic",
    "ISO8859-5",
    NULL, NULL,
    0, 0, 0, 0, 0, 0
  }
  ,
/*
   {
   "ara",
   "ISO8859-6",
   NULL, NULL,
   0, 0, 0, 0, 0, 0
   }
   ,
 */
  {
    "greek",
    "ISO8859-7",
    NULL, NULL,
    0, 0, 0, 0, 0, 0
  }
  ,
  {
    "hebrew",
    "ISO8859-8",
    NULL, NULL,
    0, 0, 0, 0, 0, 0
  }
  ,
  {
    "ltn5",
    "ISO8859-9",
    NULL, NULL,
    0, 0, 0, 0, 0, 0
  }
  ,
  {
    "viet",
    "VISCII1.1",
    NULL, NULL,
    0, 0, 0, 0, 0, 0
  }
  ,
  {
    "ethio",
    "ETHIO-0",
    NULL, NULL,
    0, 0, 0, 0, 0, 0
  }
  ,
  {
    "thai",
    "TIS620.2529",
    NULL, NULL,
    0, 0, 0, 0, 0, 0
  }
  ,
  {
    "koi8",
    "GOST19768",
    NULL, NULL,
    0, 0, 0, 0, 0, 0
  }
  ,
  {
    "arab1",
    "MULEARABIC-1",
    NULL, NULL,
    0, 0, 0, 0, 0, 0
  }
  ,
  {
    "arab2",
    "MULEARABIC-2",
    NULL, NULL,
    0, 0, 0, 0, 0, 0
  }
};


char *
getcol (row, numcol)
     char *row;
     int numcol;
{
  int i, j, k;
  char *tmpcol;
  if (numcol <= 0)
    return row;
  tmpcol = curcol;
  for (i = 0, j = 0, k = 0; row[i] != '\n'; i++)
    {
      if (row[i] == ' ' || row[i] == '\t')
	{
	  if (j)
	    {
	      tmpcol[j] = '\0';
	      j = 0;
	      if (++k == numcol)
		return tmpcol;
	    }
	  continue;
	}
      tmpcol[j++] = row[i];
    }
  if (k == numcol - 1)
    {
      tmpcol[j] = '\0';
    }
  else
    tmpcol[0] = '\0';
  return tmpcol;
}

char *
strcsstr (str1, str2)
     char *str1, *str2;
{
  int i, j;

  for (i = 0; i < (int) (strlen (str1) - strlen (str2) + 1); i++)
    {
      for (j = 0; j < strlen (str2); j++)
	{
	  /* case insensitive (not exact) */
	  if ((str2[j] & 0xDF) == (str1[i + j] & 0xDF))
	    {
	      if (j == strlen (str2) - 1)
		return str1 + i;
	    }
	  else
	    break;
	}
    }
  return 0;
}

mirror (src, nbytes)
     char *src;
{
  int bmrow, dst, i;
  sscanf (src, "%x", &bmrow);
  dst = 0;
  for (i = 0; i < nbytes * 8; i++)
    {
      dst <<= 1;
      dst |= (bmrow & 1);
      bmrow >>= 1;
    }
  return dst;
}

main (ac, av)
     char *av[];
{
  unsigned int XBMPOS, NOFCH, NOFB, CHAR, ROW, KINROW, BMB, cnvtb, x;
  int width, height, height_body, height_bottom, height_bottom_min, height_bottom_max;
  int ccnum, initc, isAscii, wbytes, fc, i, j;
  unsigned int first, last;
  char *row, charset[256], fntfile[256];
  FILE *dstf, *srcf, *infof, *fntf;
  int tabsize = 0;

  if (ac < 2)
    {
      fprintf (stderr, "Usage: %s  <x16 or x14 bdf-file-name>\n", av[0]);
      exit (1);
    }

  row = currow;
  printf ("Creating \"_xbm*.c\"...\n");

  dstf = fopen ("_xbmraw.c", "w+");
  if (dstf == NULL)
    {
      fprintf (stderr, "Error opening \"_xbmraw.c\"!");
      exit (1);
    }

  infof = fopen (info, "w+");
  if (infof == NULL)
    {
      fprintf (stderr, "Error opening \"%s\"!", info);
      exit (1);
    }

  printf ("Writing info to \"%s\"...\n", info);



  fprintf (infof, "\nList of processed fonts:\n\n");
  fprintf (infof, "%25s%15s%18s%7s%7s%7s\n",
	   "BDF File", "Name", "Register Name", "Width", "Height", "Chars");

  fprintf (dstf, "#include \"cii.h\"\n");






  for (fc = 1; fc < ac; fc++)
    {
      if (av[fc][0] == '+')
	{
	  tabsize = atoi (&av[fc][1]);
	  continue;
	}

      initc = XBMPOS = BMB = isAscii = 0;
      height = height_body = height_bottom = width = height_bottom_min = 0;
      height_bottom_max = -9999;
      fntf = 0;

      for (x = 0; x < 0xffff; x++)
	KANIND[x] = -1;

      srcf = fopen (av[fc], "r");
      if (srcf == NULL)
	{
	  fprintf (stderr, "Error opening \"%s\"!", av[fc]);
	  exit (1);
	}
      printf ("Processing %s\n", av[fc]);

      /* scanning font size */
      for (; fgets (row, 2048, srcf) != NULL;)
	{
	  if (!strcmp ("ENCODING", getcol (row, 1)))
	    {
	      int code = atoi (getcol (row, 2));
	      if (code <= 0x7F)
		continue;
	    }
	  if (!strcmp ("BBX", getcol (row, 1)))
	    {
	      int w, h, hb;
	      w = atoi (getcol (row, 2));
	      h = atoi (getcol (row, 3));
	      hb = atoi (getcol (row, 5));
	      if (hb < height_bottom_min)
		height_bottom_min = hb;
	      if (hb > height_bottom_max)
		height_bottom_max = hb;
	      if (hb > 0)
		h += hb;
	      if (height == 0)
		width = w;
	      if (height < h)
		height = h;
	      if (w != width)
		{
		  printf ("width is not fixed!!!\n");
		  goto newfont;
		}
	    }
	}

      if (height_bottom_max < 0)
	height_bottom_min -= height_bottom_max;
      height -= height_bottom_min;
      if (width > 32 || height > 32)
	{
	  printf ("font size is too large!!!\n");
	  goto newfont;
	}

      wbytes = (width + 7) / 8;
      rewind (srcf);
      for (; fgets (row, 2048, srcf) != NULL;)
	{
	  if (!fntf)
	    {
	      if (!strcmp ("FONTBOUNDINGBOX", getcol (row, 1)))
		{
		  initc |= 1;
		}
	      if (!strcmp ("CHARS", getcol (row, 1)))
		{
		  NOFCH = atoi (getcol (row, 2));
		  initc |= 2;
		}
	      if (!strcmp ("FONT", getcol (row, 1)))
		{
		  strncpy (charset, row + 5, sizeof (charset) - 1);
		  initc |= 4;
		}

	    }

	  if (!strcmp ("ENCODING", getcol (row, 1)))
	    {
	      if (!fntf && initc != 7)
		{
		  printf ("\nSomething wrong! Please check your BDF file!\n", av[fc]);
		  goto newfont;
		}

	      if (!fntf)
		{
		  for (i = 0, ccnum = -1; i < sizeof (SupChC) / sizeof (struct CharCodesXBM); i++)
		    if (strcsstr (charset, SupChC[i].regname))
		      {
			ccnum = i;
			break;
		      }

		  if (ccnum < 0)
		    {
		      printf ("charset not supported!\n");
		      goto newfont;
		    }
		  else
		    {
		      for (i = 0; i <= ChCMax; i++)
			{
			  if (ChC[i].regname == SupChC[ccnum].regname
			      && ChC[i].height == height)
			    {
			      printf ("already exists!\n");
			      goto newfont;
			      break;
			    }
			}
		      if (i > ChCMax)
			{
			  ChCMax++;
			  memcpy (&ChC[ChCMax], &SupChC[ccnum], sizeof (struct CharCodesXBM));
			}
		    }

		  strncpy (charset, ChC[ChCMax].name, 5);
		  charset[5] = 0;
		  sprintf (fntfile, "_xbm_%s_%d.c", charset, height);
		  fntf = fopen (fntfile, "w+");
		  if (fntf == NULL)
		    {
		      fprintf (stderr, "Error opening \"%s\"!", fntfile);
		      exit (1);
		    }
		  fprintf (fntf, "/* Font %s */\nchar *%s_%d_src[]={\n",
			   ChC[ChCMax].regname, ChC[ChCMax].name, height);
		}

	      NOFB = 0;
	      CHAR = (unsigned int) atoi (getcol (row, 2));

	      /* charcode conversion for ethiocode */
	      if (strcmp (ChC[ChCMax].name, "ethio") == 0)
		CHAR = ((CHAR / 94) + 0xA1) * 256 + ((CHAR % 94) + 0xA1);

	      isAscii = ((CHAR < 0x80) && strncmp (ChC[ChCMax].name, "arab", 4) != 0);

	      if (isAscii)
		NOFCH--;
	      else
		{
		  if (!XBMPOS)
		    first = CHAR;
		  last = CHAR;
		  fprintf (fntf, "/* %x */\n\"\\\n", CHAR);
		  KANIND[(unsigned int) (CHAR - first)] = XBMPOS++;
		}
	    }

	  if (!isAscii && !strcmp ("BBX", getcol (row, 1)))
	    {
	      height_body = atoi (getcol (row, 3));
	      height_bottom = atoi (getcol (row, 5)) - height_bottom_min;
	      if (height_bottom < 0)
		height_bottom = 0;
	      for (i = 0; i < height - height_body - height_bottom; i++)
		{
		  for (j = 0; j < wbytes; j++)
		    fprintf (fntf, "\\000");
		  fprintf (fntf, "\\\n");
		}
	    }

	  if (!isAscii && !strcmp ("ENDCHAR", getcol (row, 1)))
	    {
	      for (i = 0; i < height_bottom; i++)
		{
		  fprintf (fntf, "\\\n");
		  for (j = 0; j < wbytes; j++)
		    fprintf (fntf, "\\000");
		}

	      BMB = 0;
	      fprintf (fntf, "\",\n");
	      if (XBMPOS % 500 == 0 || XBMPOS == NOFCH)
		{
		  printf ("%u%%\r", (XBMPOS * 100) / NOFCH);
		  fflush (stdout);
		}
	    }


	  if (!isAscii && BMB)
	    {
	      if (NOFB)
		{
		  fprintf (fntf, "\\\n");
		}
	      cnvtb = mirror (getcol (row, 1), wbytes);

	      for (i = 0; i < wbytes; i++)
		{
		  fprintf (fntf, "\\%03o", cnvtb & 0xFF);
		  cnvtb >>= 8;
		}
	      NOFB = 1;
	    }

	  if (!isAscii && !strcmp ("BITMAP", getcol (row, 1)))
	    BMB = 1;

	}

      fprintf (fntf, "0};\n");
      fprintf (dstf, "extern char *%s_%d_src[];\n", ChC[ChCMax].name, height);
      fprintf (dstf, "extern int %s_%d_ind[];\n", ChC[ChCMax].name, height);
      fprintf (fntf, "int %s_%d_ind[]={\n", ChC[ChCMax].name, height);
      for (x = 0; x < last - first + 1; x++)
	{
	  fprintf (fntf, "%d,", KANIND[x]);
	  if (x % 10 == 9)
	    fprintf (fntf, "\n");
	}
      fprintf (fntf, "0};\n");
      ChC[ChCMax].width = width;
      ChC[ChCMax].height = height;
      ChC[ChCMax].first = first;
      ChC[ChCMax].last = last;
      ChC[ChCMax].is2bytes = (last > 0xFF) ? 1 : 0;
      ChC[ChCMax].has8bit = (last & 0x8080) ? 1 : 0;
      if (NOFCH != XBMPOS)
	{
	  fprintf (stderr, "\nSomething wrong! Please check your BDF file!\n", av[fc]);
	  exit (1);
	}
      fprintf (infof, "%25s%15s%18s%7u%7u%7u\n", strrchr (av[fc], '/') + 1, ChC[ChCMax].name, ChC[ChCMax].regname, width, height, NOFCH);
      printf ("Wrote %d chars to \"%s\"\n", XBMPOS, fntfile);
    newfont:
      fclose (fntf);
      fclose (srcf);
    }
  fclose (infof);
  fprintf (dstf, "int CII_CC_SIZE = %d;\n", ChCMax + 2 + tabsize);
  fprintf (dstf, "struct CharCodesXBM CII_CC[%d] ={\n", ChCMax + 2 + tabsize);

  for (i = 0; i <= ChCMax; i++)
    {
      fprintf (dstf, "{\"%s\",\"%s\",\n", ChC[i].name, ChC[i].regname);
      fprintf (dstf, "%s_%d_src,%s_%d_ind,\n",
	       ChC[i].name, ChC[i].height, ChC[i].name, ChC[i].height);
      fprintf (dstf, "%d,%d,\n", ChC[i].width, ChC[i].height);
      fprintf (dstf, "%u,%u,\n", ChC[i].first, ChC[i].last);
      fprintf (dstf, "%d,%d},\n", ChC[i].is2bytes, ChC[i].has8bit);
    }
  fprintf (dstf, "0};\n");
  fclose (dstf);
  printf ("\nDone with creating \"_xbm*.c\".\n");
  exit (0);
}
